import { randomBytes } from '@noble/ciphers/webcrypto';
import * as nostrTools from '@nostr/tools';
import { Database } from 'jsr:@db/sqlite';
import { NSchema as n } from 'jsr:@nostrify/nostrify';
import { encodeBase64 } from 'jsr:@std/encoding@0.224/base64';
import { UserConnection } from './UserConnection.ts';
import { handleCCNCommands } from './commands/ccn.ts';
import { handleClose } from './commands/close.ts';
import { handleEvent } from './commands/event.ts';
import { handleRequest } from './commands/request.ts';
import { CHUNK_CLEANUP_INTERVAL, QUEUE_PROCESS_INTERVAL } from './consts.ts';
import { addEventToDb } from './dbEvents/addEventToDb.ts';
import {
  ChunkedEventReceived,
  EventAlreadyExistsException,
  decryptEvent,
} from './eventEncryptionDecryption.ts';
import { runMigrations } from './migrations.ts';
import { pool, relays } from './relays.ts';
import { cleanupOldChunks } from './utils/cleanupOldChunks.ts';
import { getEveFilePath } from './utils/files.ts';
import { getEncryptedEventByOriginalId } from './utils/getEncryptedEventByOriginalId.ts';
import { isArray } from './utils/isArray.ts';
import { isLocalhost } from './utils/isLocalhost.ts';
import { isValidJSON } from './utils/isValidJSON.ts';
import {
  isOriginalEventIdCached,
  updateKnownEventsCache,
} from './utils/knownEventsCache.ts';
import { log, setupLogger } from './utils/logs.ts';
import {
  getQueueStats,
  processOutboundQueue,
  processStartupQueue,
} from './utils/outboundQueue.ts';
import { sql } from './utils/queries.ts';
import {
  SecurityEventType,
  SecuritySeverity,
  logSecurityEvent,
  logSystemEvent,
} from './utils/securityLogs.ts';

await setupLogger(null);
await Deno.mkdir(await getEveFilePath('ccn_keys'), { recursive: true });

logSystemEvent(SecurityEventType.SYSTEM_STARTUP, {
  version: 'alpha',
  deno_version: Deno.version.deno,
  timestamp: new Date().toISOString(),
});

if (!Deno.env.has('ENCRYPTION_KEY')) {
  const newKey = encodeBase64(randomBytes(32));
  log.error(
    `Missing ENCRYPTION_KEY. Please set it in your env.\nA new one has been generated for you: ENCRYPTION_KEY="${newKey}"`,
  );

  logSecurityEvent({
    eventType: SecurityEventType.CONFIGURATION_LOADED,
    severity: SecuritySeverity.CRITICAL,
    source: 'startup',
    details: {
      error: 'missing_encryption_key',
      generated_new_key: true,
    },
  });

  Deno.exit(1);
}

logSystemEvent(SecurityEventType.CONFIGURATION_LOADED, {
  encryption_key_present: true,
  ccn_keys_directory: await getEveFilePath('ccn_keys'),
});

export const db = new Database(await getEveFilePath('db'));
await setupLogger(db);

/**
 * Creates a subscription event handler for processing encrypted events.
 * This handler decrypts and adds valid events to the database.
 * @param database The database instance to use
 * @returns An event handler function
 */
function createSubscriptionEventHandler(db: Database) {
  return async (event: nostrTools.Event) => {
    if (isOriginalEventIdCached(event.id)) return;
    if (!nostrTools.verifyEvent(event)) {
      log.warn('Invalid event received');

      logSecurityEvent({
        eventType: SecurityEventType.INVALID_SIGNATURE,
        severity: SecuritySeverity.MEDIUM,
        source: 'event_processing',
        details: {
          event_id: event.id,
          event_kind: event.kind,
          pubkey: event.pubkey,
        },
      });

      return;
    }
    if (getEncryptedEventByOriginalId(db, event)) return;
    try {
      const decryptedEvent = await decryptEvent(db, event);
      addEventToDb(db, decryptedEvent, event, decryptedEvent.ccnPubkey);
      updateKnownEventsCache();
    } catch (e) {
      if (e instanceof EventAlreadyExistsException) {
        logSecurityEvent({
          eventType: SecurityEventType.DUPLICATE_EVENT_BLOCKED,
          severity: SecuritySeverity.LOW,
          source: 'event_processing',
          details: {
            event_id: event.id,
            reason: 'event_already_exists',
          },
        });
        return;
      }
      if (e instanceof ChunkedEventReceived) {
        logSecurityEvent({
          eventType: SecurityEventType.CHUNKED_EVENT_RECEIVED,
          severity: SecuritySeverity.LOW,
          source: 'event_processing',
          details: {
            event_id: event.id,
            event_kind: event.kind,
          },
        });
        return;
      }

      logSecurityEvent({
        eventType: SecurityEventType.DECRYPTION_FAILURE,
        severity: SecuritySeverity.MEDIUM,
        source: 'event_processing',
        details: {
          operation: 'event_decryption',
          event_id: event.id,
          event_kind: event.kind,
          error_message: e instanceof Error ? e.message : 'Unknown error',
        },
      });
    }
  };
}

function setupAndSubscribeToExternalEvents() {
  const isInitialized = sql`
    SELECT name FROM sqlite_master WHERE type='table' AND name='migration_history'
  `(db)[0];

  if (!isInitialized) runMigrations(db, -1);

  const latestVersion =
    sql`
      SELECT migration_version FROM migration_history WHERE status = 'success' ORDER BY migration_version DESC LIMIT 1
    `(db)[0]?.migration_version ?? -1;

  runMigrations(db, latestVersion);

  logSystemEvent(SecurityEventType.SYSTEM_STARTUP, {
    database_logging_enabled: true,
    migrations_complete: true,
  });

  const allCCNs = sql`SELECT pubkey FROM ccns`(db);
  const ccnPubkeys = allCCNs.map((ccn) => ccn.pubkey);

  pool.subscribeMany(
    relays,
    [
      {
        '#p': ccnPubkeys,
        kinds: [1059],
      },
    ],
    {
      onevent: createSubscriptionEventHandler(db),
    },
  );

  updateKnownEventsCache();
  setInterval(() => cleanupOldChunks(db), CHUNK_CLEANUP_INTERVAL);

  processStartupQueue(db).catch((error: unknown) => {
    log.error('Startup queue processing failed', {
      tag: 'outboundQueue',
      error: error instanceof Error ? error.message : 'Unknown error',
    });
  });

  setInterval(async () => {
    try {
      await processOutboundQueue(db);

      const stats = getQueueStats(db);
      if (stats.total > 0) {
        log.info('Outbound queue status', {
          tag: 'outboundQueue',
          ...stats,
        });
      }
    } catch (error) {
      log.error('Error processing outbound queue', {
        tag: 'outboundQueue',
        error: error instanceof Error ? error.message : 'Unknown error',
      });
    }
  }, QUEUE_PROCESS_INTERVAL);

  processOutboundQueue(db).catch((error) => {
    log.error('Initial queue processing failed', {
      tag: 'outboundQueue',
      error: error instanceof Error ? error.message : 'Unknown error',
    });
  });

  log.info('Outbound queue processor started', {
    tag: 'outboundQueue',
    interval: QUEUE_PROCESS_INTERVAL,
  });
}

setupAndSubscribeToExternalEvents();

export function handleSocketError(
  connection: UserConnection,
  operation: string,
  error: unknown,
): void {
  const errorMessage = error instanceof Error ? error.message : 'Unknown error';
  log.error(`Error ${operation}: ${errorMessage}`);
  connection.sendNotice(`Failed to ${operation}`);
}

Deno.serve({
  port: 6942,
  handler: (request, connInfo) => {
    if (request.headers.get('upgrade') === 'websocket') {
      if (!isLocalhost(request, connInfo)) {
        logSecurityEvent({
          eventType: SecurityEventType.NON_LOCALHOST_CONNECTION_BLOCKED,
          severity: SecuritySeverity.HIGH,
          source: 'connection_security',
          details: {
            remote_addr: connInfo?.remoteAddr,
            user_agent: request.headers.get('user-agent'),
            host: request.headers.get('host'),
            origin: request.headers.get('origin'),
          },
          remoteAddr:
            connInfo?.remoteAddr?.transport === 'tcp'
              ? connInfo.remoteAddr.hostname
              : 'unknown',
        });

        return new Response(
          'Forbidden. Please read the Arx-CCN documentation for more information on how to interact with the relay.',
          { status: 403 },
        );
      }

      log.info('upgrading connection', { tag: 'WebSocket' });
      const { socket, response } = Deno.upgradeWebSocket(request);

      const connection = new UserConnection(socket, new Map(), db);

      socket.onopen = () => {
        log.info('User connected');

        logSecurityEvent({
          eventType: SecurityEventType.WEBSOCKET_CONNECTION_ESTABLISHED,
          severity: SecuritySeverity.LOW,
          source: 'connection_security',
          details: {
            remote_addr: connInfo?.remoteAddr,
            user_agent: request.headers.get('user-agent'),
            connection_time: new Date().toISOString(),
          },
          remoteAddr:
            connInfo?.remoteAddr?.transport === 'tcp'
              ? connInfo.remoteAddr.hostname
              : 'localhost',
        });
      };

      socket.onmessage = (event) => {
        log.debug('Received', {
          tag: 'WebSocket',
          data: event.data,
        });

        if (typeof event.data !== 'string' || !isValidJSON(event.data)) {
          log.warn('Invalid request', { tag: 'WebSocket' });

          logSecurityEvent({
            eventType: SecurityEventType.MALFORMED_EVENT,
            severity: SecuritySeverity.MEDIUM,
            source: 'websocket_handler',
            details: {
              data_type: typeof event.data,
              is_valid_json: isValidJSON(event.data),
              data_length: event.data?.length || 0,
            },
            remoteAddr:
              connInfo?.remoteAddr?.transport === 'tcp'
                ? connInfo.remoteAddr.hostname
                : 'localhost',
          });

          return;
        }

        const data = JSON.parse(event.data);
        if (!isArray(data)) {
          logSecurityEvent({
            eventType: SecurityEventType.MALFORMED_EVENT,
            severity: SecuritySeverity.MEDIUM,
            source: 'websocket_handler',
            details: {
              error: 'message_not_array',
              received_type: typeof data,
            },
            remoteAddr:
              connInfo?.remoteAddr?.transport === 'tcp'
                ? connInfo.remoteAddr.hostname
                : 'localhost',
          });

          return log.warn('Invalid request', { tag: 'WebSocket' });
        }

        const msgType = data[0];
        log.debug('received message', { tag: 'WebSocket', msgType });

        switch (msgType) {
          case 'REQ':
            return handleRequest(connection, n.clientREQ().parse(data));
          case 'EVENT':
            return handleEvent(connection, n.clientEVENT().parse(data)[1]);
          case 'CLOSE':
            return handleClose(connection, n.clientCLOSE().parse(data)[1]);
          case 'CCN':
            return handleCCNCommands(connection, data[1] as string, data[2]);
          default:
            logSecurityEvent({
              eventType: SecurityEventType.MALFORMED_EVENT,
              severity: SecuritySeverity.MEDIUM,
              source: 'websocket_handler',
              details: {
                error: 'unknown_message_type',
                message_type: msgType,
                data_preview: data.slice(0, 3), // First 3 elements for debugging
              },
              remoteAddr:
                connInfo?.remoteAddr?.transport === 'tcp'
                  ? connInfo.remoteAddr.hostname
                  : 'localhost',
            });

            log.warn('Invalid request', { tag: 'WebSocket' });
            return;
        }
      };

      socket.onclose = () => {
        log.info('User disconnected');

        logSecurityEvent({
          eventType: SecurityEventType.WEBSOCKET_CONNECTION_CLOSED,
          severity: SecuritySeverity.LOW,
          source: 'connection_security',
          details: {
            disconnect_time: new Date().toISOString(),
            subscriptions_count: connection.subscriptions.size,
          },
          remoteAddr:
            connInfo?.remoteAddr?.transport === 'tcp'
              ? connInfo.remoteAddr.hostname
              : 'localhost',
        });
      };

      return response;
    }
    return new Response(
      Deno.readTextFileSync(`${import.meta.dirname}/../public/landing.html`),
      {
        headers: { 'Content-Type': 'text/html' },
      },
    );
  },
});
