import type { Database } from '@db/sqlite';
import { log } from './utils/logs.ts';
import { sql } from './utils/queries.ts';

export function runMigrations(db: Database, latestVersion: number) {
  const migrations = [
    ...Deno.readDirSync(`${import.meta.dirname}/../migrations`),
  ];
  migrations.sort((a, b) => {
    const aVersion = Number.parseInt(a.name.split('-')[0], 10);
    const bVersion = Number.parseInt(b.name.split('-')[0], 10);
    return aVersion - bVersion;
  });
  for (const migrationFile of migrations) {
    const migrationVersion = Number.parseInt(
      migrationFile.name.split('-')[0],
      10,
    );

    if (migrationVersion > latestVersion) {
      log.info(
        `Running migration ${migrationFile.name} (version ${migrationVersion})`,
      );
      const start = Date.now();
      const migrationSql = Deno.readTextFileSync(
        `${import.meta.dirname}/../migrations/${migrationFile.name}`,
      );
      db.run('BEGIN TRANSACTION');
      try {
        db.run(migrationSql);
        const end = Date.now();
        const durationMs = end - start;
        sql`
          INSERT INTO migration_history (migration_version, migration_name, executed_at, duration_ms, status) VALUES (${migrationVersion}, ${migrationFile.name}, ${new Date().toISOString()}, ${durationMs}, 'success');
          db.run("COMMIT TRANSACTION");
        `(db);
      } catch (e) {
        db.run('ROLLBACK TRANSACTION');
        const error =
          e instanceof Error
            ? e
            : typeof e === 'string'
              ? new Error(e)
              : new Error(JSON.stringify(e));
        const end = Date.now();
        const durationMs = end - start;
        sql`
          INSERT INTO migration_history (migration_version, migration_name, executed_at, duration_ms, status, error_message) VALUES (${migrationVersion}, ${migrationFile.name}, ${new Date().toISOString()}, ${durationMs}, 'failed', ${error.message});
        `(db);
        throw e;
      }
      db.run('END TRANSACTION');
    }
  }
}
