import { encodeBase64 } from 'jsr:@std/encoding@~0.224.1/base64';
import type { Database } from '@db/sqlite';
import * as nostrTools from '@nostr/tools';
import * as nip06 from '@nostr/tools/nip06';
import { encryptUint8Array, encryptionKey } from './encryption.ts';
import { getEveFilePath } from './files.ts';
import { sql } from './queries.ts';

/**
 * Create a new CCN and store it in the database
 *
 * @param db - The database instance
 * @param name - The name of the CCN
 * @param seed - The seed words for the CCN
 * @returns The public key and private key of the CCN
 */

export async function createNewCCN(
  db: Database,
  name: string,
  creator: string,
  seed?: string,
): Promise<{ pubkey: string; privkey: Uint8Array }> {
  const ccnSeed = seed || nip06.generateSeedWords();
  const ccnPrivateKey = nip06.privateKeyFromSeedWords(ccnSeed);
  const ccnPublicKey = nostrTools.getPublicKey(ccnPrivateKey);

  const ccnSeedPath = await getEveFilePath(`ccn_seeds/${ccnPublicKey}`);
  const ccnPrivPath = await getEveFilePath(`ccn_keys/${ccnPublicKey}`);

  await Deno.mkdir(await getEveFilePath('ccn_seeds'), { recursive: true });
  await Deno.mkdir(await getEveFilePath('ccn_keys'), { recursive: true });

  const encryptedPrivateKey = encryptUint8Array(ccnPrivateKey, encryptionKey);

  Deno.writeTextFileSync(ccnSeedPath, ccnSeed);
  Deno.writeTextFileSync(ccnPrivPath, encodeBase64(encryptedPrivateKey));

  db.run('BEGIN TRANSACTION');

  sql`INSERT INTO ccns (pubkey, name) VALUES (${ccnPublicKey}, ${name})`(db);
  sql`INSERT INTO allowed_writes (ccn_pubkey, pubkey) VALUES (${ccnPublicKey}, ${creator})`(
    db,
  );

  db.run('COMMIT TRANSACTION');

  return {
    pubkey: ccnPublicKey,
    privkey: ccnPrivateKey,
  };
}
