import type { Database, Statement } from '@db/sqlite';
import * as log from '@std/log';

interface DatabaseHandlerOptions extends log.BaseHandlerOptions {
  db?: Database;
}

export class DatabaseHandler extends log.BaseHandler {
  private db: Database | null = null;
  private insertStmt: Statement | null = null;

  constructor(levelName: log.LevelName, options: DatabaseHandlerOptions = {}) {
    super(levelName, options);
    if (options.db) {
      this.setDatabase(options.db);
    }
  }

  setDatabase(db: Database): void {
    this.db = db;
    try {
      this.insertStmt = this.db.prepare(`
        INSERT INTO logs (timestamp, level, message, args, source, event_type, severity, remote_addr, ccn_pubkey, event_id, risk_score)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
      `);
    } catch (error) {
      console.warn('Database logger not ready:', error);
    }
  }

  override log(msg: string): void {
    // Required by the abstract base class but not used in our implementation
  }

  override handle(logRecord: log.LogRecord): void {
    if (this.shouldSkipLogging(logRecord)) {
      return;
    }

    if (!this.db || !this.insertStmt) {
      return;
    }

    try {
      const timestamp = new Date(logRecord.datetime).toISOString();
      const level = this.getLevelName(logRecord.level);
      const message = logRecord.msg;

      const securityData = this.extractSecurityData(logRecord.args);
      const sanitizedArgs = this.sanitizeArgs(logRecord.args);
      const argsJson =
        sanitizedArgs.length > 0 ? JSON.stringify(sanitizedArgs) : null;
      const source = this.extractSource(logRecord.args);

      this.insertStmt.run(
        timestamp,
        level,
        message,
        argsJson,
        source,
        securityData.eventType,
        securityData.severity,
        securityData.remoteAddr,
        securityData.ccnPubkey,
        securityData.eventId,
        securityData.riskScore,
      );
    } catch (error) {
      console.error('Failed to write log to database:', error);
    }
  }

  private getLevelName(level: number): string {
    switch (level) {
      case 10:
        return 'DEBUG';
      case 20:
        return 'INFO';
      case 30:
        return 'WARN';
      case 40:
        return 'ERROR';
      case 50:
        return 'FATAL';
      default:
        return `LVL${level}`;
    }
  }

  private shouldSkipLogging(logRecord: log.LogRecord): boolean {
    const message = logRecord.msg.toLowerCase();

    if (
      message.includes('sql') ||
      message.includes('database') ||
      message.includes('migration') ||
      message.includes('sqlite')
    ) {
      return true;
    }

    if (message.includes('log') && message.includes('database')) {
      return true;
    }

    return false;
  }

  private extractSecurityData(args: unknown[]): {
    eventType: string | null;
    severity: string | null;
    remoteAddr: string | null;
    ccnPubkey: string | null;
    eventId: string | null;
    riskScore: number | null;
  } {
    let eventType = null;
    let severity = null;
    let remoteAddr = null;
    let ccnPubkey = null;
    let eventId = null;
    let riskScore = null;

    for (const arg of args) {
      if (typeof arg === 'object' && arg !== null) {
        const obj = arg as Record<string, unknown>;

        if (obj.eventType && typeof obj.eventType === 'string') {
          eventType = obj.eventType;
        }
        if (obj.severity && typeof obj.severity === 'string') {
          severity = obj.severity;
        }
        if (obj.remoteAddr && typeof obj.remoteAddr === 'string') {
          remoteAddr = obj.remoteAddr;
        }
        if (obj.ccnPubkey && typeof obj.ccnPubkey === 'string') {
          ccnPubkey = obj.ccnPubkey;
        }
        if (obj.eventId && typeof obj.eventId === 'string') {
          eventId = obj.eventId;
        }
        if (obj.risk_score && typeof obj.risk_score === 'number') {
          riskScore = obj.risk_score;
        }
      }
    }

    return { eventType, severity, remoteAddr, ccnPubkey, eventId, riskScore };
  }

  private extractSource(args: unknown[]): string | null {
    for (const arg of args) {
      if (typeof arg === 'object' && arg !== null) {
        const obj = arg as Record<string, unknown>;
        if (obj.tag && typeof obj.tag === 'string') {
          return obj.tag;
        }
        if (obj.source && typeof obj.source === 'string') {
          return obj.source;
        }
      }
    }
    return null;
  }

  private sanitizeArgs(args: unknown[]): unknown[] {
    const sensitiveKeys = [
      'privatekey',
      'private_key',
      'privkey',
      'priv_key',
      'secretkey',
      'secret_key',
      'seckey',
      'sec_key',
      'password',
      'pass',
      'pwd',
      'token',
      'auth',
      'ccnprivatekey',
      'ccn_private_key',
      'ccnprivkey',
    ];

    return args.map((arg) => {
      if (typeof arg === 'object' && arg !== null) {
        const sanitized: Record<string, unknown> = {};

        for (const [key, value] of Object.entries(
          arg as Record<string, unknown>,
        )) {
          const lowerKey = key.toLowerCase();

          if (
            sensitiveKeys.some((sensitiveKey) =>
              lowerKey.includes(sensitiveKey),
            )
          ) {
            sanitized[key] = '[REDACTED]';
          } else if (value instanceof Uint8Array) {
            sanitized[key] = `[Uint8Array length=${value.length}]`;
          } else {
            sanitized[key] = value;
          }
        }

        return sanitized;
      }

      return arg;
    });
  }
}

export const dbHandler: DatabaseHandler | null = null;
