import type * as nostrTools from '@nostr/tools';
import type { Database } from 'jsr:@db/sqlite';
import { MAX_TRANSMISSION_ATTEMPTS } from '../consts.ts';
import { publishToRelays, relays } from '../relays.ts';
import { log } from './logs.ts';
import { sql } from './queries.ts';
import {
  SecurityEventType,
  SecuritySeverity,
  logSecurityEvent,
} from './securityLogs.ts';

export interface QueuedEvent {
  queue_id: number;
  event_id: string;
  encrypted_event: string;
  ccn_pubkey: string;
  created_at: number;
  attempts: number;
  last_attempt: number | null;
  status: 'pending' | 'sending' | 'sent' | 'failed' | 'stale';
  error_message: string | null;
}

let isConnectedToRelays = false;
let lastSuccessfulTransmission = 0;
let consecutiveFailures = 0;

export function queueEventForTransmission(
  db: Database,
  eventId: string,
  encryptedEvent: nostrTools.VerifiedEvent | nostrTools.VerifiedEvent[],
  ccnPubkey: string,
): void {
  try {
    const encryptedEventJson = JSON.stringify(encryptedEvent);

    sql`
      INSERT OR REPLACE INTO outbound_event_queue
      (event_id, encrypted_event, ccn_pubkey, status)
      VALUES (${eventId}, ${encryptedEventJson}, ${ccnPubkey}, 'pending')
    `(db);

    log.debug('Event queued for transmission', {
      tag: 'outboundQueue',
      eventId,
      ccnPubkey,
      connectionState: isConnectedToRelays ? 'connected' : 'offline',
    });

    logSecurityEvent({
      eventType: SecurityEventType.EVENT_QUEUED_FOR_TRANSMISSION,
      severity: SecuritySeverity.LOW,
      source: 'outbound_queue',
      details: {
        action: 'event_queued',
        event_id: eventId,
        ccn_pubkey: ccnPubkey,
        is_chunked: Array.isArray(encryptedEvent),
        connection_state: isConnectedToRelays,
      },
    });
  } catch (error) {
    log.error('Failed to queue event for transmission', {
      tag: 'outboundQueue',
      eventId,
      error,
    });
    throw error;
  }
}

export function getPendingEvents(db: Database, limit = 50): QueuedEvent[] {
  const now = Math.floor(Date.now() / 1000);
  const maxAge = 30 * 24 * 60 * 60;

  sql`
    UPDATE outbound_event_queue
    SET status = 'stale'
    WHERE status IN ('pending', 'failed')
    AND created_at < ${now - maxAge}
  `(db);

  return sql`
    SELECT * FROM outbound_event_queue
    WHERE (
      status = 'pending'
      OR (
        status = 'failed'
        AND attempts < ${MAX_TRANSMISSION_ATTEMPTS}
        AND (
          last_attempt IS NULL OR
          last_attempt < ${now - getAdaptiveRetryDelay(consecutiveFailures)}
        )
      )
    )
    AND status != 'stale'
    ORDER BY
      CASE WHEN status = 'pending' THEN 0 ELSE 1 END,
      created_at ASC
    LIMIT ${limit}
  `(db) as QueuedEvent[];
}

function getAdaptiveRetryDelay(failures: number): number {
  return Math.min(300, 2 ** failures * 10);
}

async function checkRelayConnectivity(): Promise<boolean> {
  const connectivityTests = relays.map(async (relay) => {
    try {
      const ws = new WebSocket(relay);

      return new Promise<boolean>((resolve) => {
        const timeout = setTimeout(() => {
          ws.close();
          resolve(false);
        }, 3000); // 3 second timeout

        ws.onopen = () => {
          clearTimeout(timeout);
          ws.close();
          resolve(true);
        };

        ws.onerror = () => {
          clearTimeout(timeout);
          resolve(false);
        };

        ws.onclose = () => {
          clearTimeout(timeout);
          resolve(false);
        };
      });
    } catch {
      return false;
    }
  });

  try {
    const results = await Promise.allSettled(connectivityTests);
    const successfulConnections = results.filter(
      (result) => result.status === 'fulfilled' && result.value === true,
    ).length;

    const isConnected = successfulConnections > 0;

    log.debug('Relay connectivity check completed', {
      tag: 'outboundQueue',
      successfulConnections,
      totalTested: relays.length,
      isConnected,
    });

    return isConnected;
  } catch {
    return false;
  }
}

export async function processStartupQueue(db: Database): Promise<void> {
  const startupEvents = sql`
    SELECT COUNT(*) as count FROM outbound_event_queue
    WHERE status IN ('pending', 'failed', 'sending')
  `(db)[0].count;

  if (startupEvents > 0) {
    log.info(`Found ${startupEvents} events from previous session`, {
      tag: 'outboundQueue',
    });

    sql`
      UPDATE outbound_event_queue
      SET status = 'failed',
          attempts = attempts + 1,
          error_message = 'Interrupted by shutdown'
      WHERE status = 'sending'
    `(db);

    await processOutboundQueue(db);
  }
}

export function markEventSending(db: Database, queueId: number): void {
  const now = Math.floor(Date.now() / 1000);

  sql`
    UPDATE outbound_event_queue
    SET status = 'sending', last_attempt = ${now}
    WHERE queue_id = ${queueId}
  `(db);
}

export function markEventSent(db: Database, queueId: number): void {
  sql`
    UPDATE outbound_event_queue
    SET status = 'sent'
    WHERE queue_id = ${queueId}
  `(db);

  isConnectedToRelays = true;
  lastSuccessfulTransmission = Math.floor(Date.now() / 1000);
  consecutiveFailures = 0;

  log.debug('Event marked as sent', {
    tag: 'outboundQueue',
    queueId,
  });
}

export function markEventFailed(
  db: Database,
  queueId: number,
  errorMessage: string,
): void {
  sql`
    UPDATE outbound_event_queue
    SET status = 'failed',
        attempts = attempts + 1,
        error_message = ${errorMessage}
    WHERE queue_id = ${queueId}
  `(db);

  consecutiveFailures++;
  const timeSinceLastSuccess =
    Math.floor(Date.now() / 1000) - lastSuccessfulTransmission;

  if (consecutiveFailures >= 3 || timeSinceLastSuccess > 300) {
    isConnectedToRelays = false;
  }

  log.warn('Event transmission failed', {
    tag: 'outboundQueue',
    queueId,
    errorMessage,
    consecutiveFailures,
    timeSinceLastSuccess,
    connectionState: isConnectedToRelays ? 'connected' : 'offline',
  });

  logSecurityEvent({
    eventType: SecurityEventType.SYSTEM_STARTUP,
    severity: SecuritySeverity.MEDIUM,
    source: 'outbound_queue',
    details: {
      action: 'transmission_failed',
      queue_id: queueId,
      error_message: errorMessage,
      consecutive_failures: consecutiveFailures,
    },
  });
}

async function batchProcessEvents(
  db: Database,
  events: QueuedEvent[],
): Promise<void> {
  const BATCH_SIZE = 10;
  const batches = [];

  for (let i = 0; i < events.length; i += BATCH_SIZE) {
    batches.push(events.slice(i, i + BATCH_SIZE));
  }

  for (const batch of batches) {
    const promises = batch.map(async (queuedEvent) => {
      try {
        const encryptedEvent = JSON.parse(queuedEvent.encrypted_event);
        await publishToRelays(encryptedEvent);
        return { queueId: queuedEvent.queue_id, success: true };
      } catch (error) {
        const errorMessage =
          error instanceof Error ? error.message : 'Unknown error';
        return {
          queueId: queuedEvent.queue_id,
          success: false,
          error: errorMessage,
        };
      }
    });

    const results = await Promise.allSettled(promises);

    results.forEach((result, index) => {
      const queuedEvent = batch[index];
      if (result.status === 'fulfilled' && result.value.success) {
        markEventSent(db, queuedEvent.queue_id);
      } else {
        const error =
          result.status === 'fulfilled'
            ? result.value.error
            : 'Promise rejected';
        markEventFailed(db, queuedEvent.queue_id, error || 'Unknown error');
      }
    });

    if (batches.indexOf(batch) < batches.length - 1) {
      await new Promise((resolve) => setTimeout(resolve, 1000));
    }
  }
}

export async function processOutboundQueue(db: Database): Promise<void> {
  const pendingEvents = getPendingEvents(db);

  if (pendingEvents.length === 0) {
    return;
  }

  const connectivityResult = await checkRelayConnectivity();
  if (!connectivityResult && consecutiveFailures > 5) {
    log.debug('Skipping queue processing - appears to be offline', {
      tag: 'outboundQueue',
      consecutiveFailures,
    });
    return;
  }

  log.info(`Processing ${pendingEvents.length} pending events`, {
    tag: 'outboundQueue',
    connectionState: isConnectedToRelays ? 'connected' : 'unknown',
  });

  for (const event of pendingEvents) {
    markEventSending(db, event.queue_id);
  }

  await batchProcessEvents(db, pendingEvents);
}

export function getQueueStats(db: Database): {
  pending: number;
  sending: number;
  sent: number;
  failed: number;
  stale: number;
  total: number;
  connectionState: string;
  consecutiveFailures: number;
} {
  const stats = sql`
    SELECT
      status,
      COUNT(*) as count
    FROM outbound_event_queue
    GROUP BY status
  `(db);

  const result = {
    pending: 0,
    sending: 0,
    sent: 0,
    failed: 0,
    stale: 0,
    total: 0,
    connectionState: isConnectedToRelays ? 'connected' : 'offline',
    consecutiveFailures,
  };

  for (const stat of stats) {
    const status = stat.status as
      | 'pending'
      | 'sending'
      | 'sent'
      | 'failed'
      | 'stale';
    switch (status) {
      case 'pending':
        result.pending = stat.count;
        break;
      case 'sending':
        result.sending = stat.count;
        break;
      case 'sent':
        result.sent = stat.count;
        break;
      case 'failed':
        result.failed = stat.count;
        break;
      case 'stale':
        result.stale = stat.count;
        break;
    }
    result.total += stat.count;
  }

  return result;
}
